@@ -1,6 +1,24 @@
 #!/usr/bin/env ruby
 
-$KCODE = 'e'
+# main.rb - samidare main program
+#
+# Copyright (C) 2003,2004,2005,2006 Tanaka Akira  <akr@fsij.org>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+#$KCODE = 'e'
 
 require 'open-uri'
 require 'uri'
@@ -17,6 +35,7 @@ require 'autofile'
 require 'htree'
 require 'string-util'
 require 'tempfile'
+require 'presen'
 
 CONFIG_FILENAME = 'config.yml'
 STATUS_FILENAME = 'status.rm'
@@ -104,23 +123,39 @@ class Entry
   end
 
   def next_timing
+    min = @config['MinimumInterval'] # xxx: should support units other than second.
+    max = @config['MaximumInterval']
+    if !min
+      if max && max < DefaultMinimumInterval
+        min = max
+      else
+        min = DefaultMinimumInterval
+      end
+    end
+    if !max
+      if min && DefaultMaximumInterval < min
+        max = min
+      else
+        max = DefaultMaximumInterval
+      end
+    end
+    max = min if max < min
     if @status['_update_info']
       Time.now
     elsif e1 = find_first_error
-      min = @config.fetch('MinimumInterval', DefaultMinimumInterval) # xxx: should support units other than second.
       e2 = find_last_error
       e1t = e1['clientDateEnd']
       e2t = e2['clientDateEnd']
       if e1.equal? e2
-        e1t + min
+        r = e1t + min
       else
-        e2t + (e2t - e1t)
+        r = e2t + (e2t - e1t)
       end
+      if r < e1t + min
+        r = e1t + min
+      end
+      r
     elsif s1 = find_first_200_with_current_content # Is there any 200?
-      min = @config.fetch('MinimumInterval', DefaultMinimumInterval) # xxx: should support units other than second.
-      max = @config.fetch('MaximumInterval', DefaultMaximumInterval)
-      min = max if max < min
-
       s2 = find_last_success
       s2t = s2['clientDateEnd'] || s2['clientDateBeg']
 
@@ -130,15 +165,21 @@ class Entry
       end
       if !s1t
         begin
-          s1t = Time.httpdate_robust(s1['lastModifiedString']) +
-            ((s1['clientDateEnd'] || s1['clientDateBeg']) - Time.httpdate_robust(s1['serverDateString']))
+          s1_client_date = s1['clientDateEnd'] || s1['clientDateBeg']
+          s1_server_date = Time.httpdate_robust(s1['serverDateString'])
+          s1t = Time.httpdate_robust(s1['lastModifiedString'])
+          s1t += s1_client_date - s1_server_date
         rescue
           s1t = s1['clientDateEnd'] || s1['clientDateBeg']
         end
         s1t = s2t if s2t < s1t
       end
 
-      r = s2t + (s2t - s1t)
+      begin
+        r = s2t + (s2t - s1t)
+      rescue RangeError
+        r = s2t + max
+      end
       if r < s2t + min
         s2t + min
       elsif s2t + max < r
@@ -159,18 +200,18 @@ class Entry
     log = @config.dup
 
     log['clientDateBeg'] = client_date_1 = Time.now
-    STDERR.puts "#{client_date_1.iso8601} fetch start #{uri}" if $VERBOSE
+    STDERR.puts "#{client_date_1.iso8601} fetch start #{uri}\n" if $VERBOSE
     page, meta = fetch(log)
     log['clientDateEnd'] = client_date_2 = Time.now
     if $VERBOSE
       if log['trouble']
-        STDERR.puts "#{client_date_2.iso8601} fetch end ERROR: #{log['trouble']} #{uri}"
+        STDERR.puts "#{client_date_2.iso8601} fetch end ERROR: #{log['trouble']} #{uri}\n"
         if log['backtrace']
-          STDERR.puts "|#{client_date_2.iso8601} ERROR: #{log['trouble']} (#{log['exception_class']}) #{uri}"
-          log['backtrace'].each {|pos| STDERR.puts "| #{pos}" }
+          STDERR.puts "|#{client_date_2.iso8601} ERROR: #{log['trouble']} (#{log['exception_class']}) #{uri}\n"
+          log['backtrace'].each {|pos| STDERR.puts "| #{pos}\n" }
         end
       else
-        STDERR.puts "#{client_date_2.iso8601} fetch end #{log['status']} #{log['statusMessage']} #{uri}"
+        STDERR.puts "#{client_date_2.iso8601} fetch end #{log['status']} #{log['statusMessage']} #{uri}\n"
       end
     end
 
@@ -179,8 +220,8 @@ class Entry
         examine(page, meta, log)
       end
     rescue
-      STDERR.puts "|#{Time.now.iso8601} examine ERROR: #{$!.message} (#{$!.class}) #{uri}"
-      $!.backtrace.each {|pos| STDERR.puts "| #{pos}" }
+      STDERR.puts "|#{Time.now.iso8601} examine ERROR: #{$!.message} (#{$!.class}) #{uri}\n"
+      $!.backtrace.each {|pos| STDERR.puts "| #{pos}\n" }
 
       log['trouble'] = "exception"
       log['backtrace'] = $!.backtrace
@@ -209,6 +250,9 @@ class Entry
       end
       opts['If-None-Match'] = h['eTag'] if h['eTag']
     end
+    if @config.include? 'Header'
+      opts.update @config['Header']
+    end
     #PP.pp([uri, opts], STDERR) if $VERBOSE
     page = nil
     meta = nil
@@ -219,6 +263,9 @@ class Entry
     backtrace = nil
     begin
       page = timeout(@config['Timeout'] || 200) { URI.parse(uri).read(opts) }
+      if page.empty?
+        trouble = "empty page"
+      end
       meta = page.meta
       status = page.status[0]
       status_message = page.status[1]
@@ -275,6 +322,16 @@ class Entry
     content_charset = content_charset.downcase
     log['contentCharset'] = content_charset if content_charset
 
+    if "".respond_to? :force_encoding
+      if content && content_charset
+        begin
+          enc = Encoding.find(content_charset)
+          content.force_encoding(enc)
+        rescue ArgumentError
+        end
+      end
+    end
+
     # checksum for gzip/deflate decoded content.
     # compression level is not affected.
     log['checksum'] = content.sum
@@ -303,7 +360,7 @@ class Entry
     t = HTree.parse(decoded_content)
 
     title = t.title
-    log['extractedTitle'] = title.to_s if title
+    log['extractedTitle'] = title.to_s.strip if title
 
     author = t.author
     log['extractedAuthor'] = author.to_s if author
@@ -326,12 +383,34 @@ class Entry
                                '{http://my.netscape.com/rdf/simple/0.9/}link')
         base_uri = URI.parse(log['baseURI'] || @config['URI'])
         link_uri = base_uri + link.extract_text.to_s.strip
-        if link_uri.scheme == 'http' && base_uri.host == link_uri.host # xxx: refine safety test
+        if link_uri.scheme == 'http' && base_uri.host.downcase == link_uri.host.downcase # xxx: refine safety test
           log['extractedLinkURI'] = link_uri.to_s
         end
       end
     end
 
+    if root
+      ['http://www.w3.org/2005/Atom', 'http://purl.org/atom/ns#'].each {|xmlns|
+        if root.name == "{#{xmlns}}feed"
+          link_uri = nil
+          t.traverse_element("{#{xmlns}}link") {|e|
+            begin
+              next unless e.fetch_attr('rel').downcase == 'alternate'
+              base_uri = URI.parse(log['baseURI'] || @config['URI'])
+              link_uri = base_uri + e.fetch_attr('href').to_s.strip
+              break
+            rescue IndexError, ArgumentError
+            end
+          }
+          if link_uri and
+              link_uri.scheme == 'http' && base_uri.host.downcase == link_uri.host.downcase # xxx: refine safety test
+            log['extractedLinkURI'] = link_uri.to_s
+            break
+          end
+        end
+      }
+    end
+
     t, checksum_filter = ignore_tree(t)
     log['checksum_filter'] = checksum_filter unless checksum_filter.empty?
     log['checksum_filtered'] = t.extract_text.rcdata.gsub(/\s+/, '').sum
@@ -357,12 +436,26 @@ class Entry
       ignore_id = []
     end
 
+    if ignore_element = config['IgnoreElement']
+      ignore_element = ignore_element.split(/\s+/)
+    else
+      ignore_element = []
+    end
+
+    ignore_text = []
+    if ignore_text_raw = config['IgnoreText']
+      ignore_text_raw.scan(%r{/((?:[^/\\]|\\.)*)/}) { ignore_text << Regexp.compile($1) }
+    end
+
     t = tree.make_loc.filter {|e|
       path = e.path.sub(%r{^doc\(\)}, '')
       e = e.to_node
       not (
+        (e.text? && (/\A\s*\z/ =~ e.to_s ||
+                     ignore_text.any? {|pat| pat =~ e.to_s })) ||
         (e.elem? && (%r[\A(?:\{http://www.w3.org/1999/xhtml\})?style\z] =~ e.name ||
-                     %r[\A(?:\{http://www.w3.org/1999/xhtml\})?script\z] =~ e.name)) ||
+                     %r[\A(?:\{http://www.w3.org/1999/xhtml\})?script\z] =~ e.name ||
+                     ignore_element.include?(e.name))) ||
         ignore_pattern =~ path ||
         (e.elem? && ((cs = e.get_attr('class') and ignore_class & cs.split(/\s+/) != []) ||
                      ignore_id.include?(e.get_attr('id'))))
@@ -374,6 +467,8 @@ class Entry
     checksum_filter.concat ['IgnorePath', *ignore_path] if !ignore_path.empty?
     checksum_filter.concat ['IgnoreClass', *ignore_class] if !ignore_class.empty?
     checksum_filter.concat ['IgnoreID', *ignore_id] if !ignore_id.empty?
+    checksum_filter.concat ['IgnoreElement', *ignore_element] if !ignore_element.empty?
+    checksum_filter.concat ['IgnoreText', *ignore_text] if !ignore_text.empty?
 
     [t, checksum_filter]
   end
@@ -391,10 +486,10 @@ class Entry
         if record.last_modified != '0' &&
            record.last_detected != '0' &&
            record.last_modified.to_i > record.last_detected.to_i
-          STDERR.puts "#{now.iso8601} info: strange LIRS: Last-Modified/Detected inversion: #{record.last_modified.to_i - record.last_detected.to_i}sec #{uri} #{record.encode.inspect}"
+          STDERR.puts "#{now.iso8601} info: strange LIRS: Last-Modified/Detected inversion: #{record.last_modified.to_i - record.last_detected.to_i}sec #{uri} #{record.encode.inspect}\n"
         end
         if record.last_modified != '0' && record.last_modified.to_i > now.to_i
-          STDERR.puts "#{now.iso8601} info: strange LIRS: future Last-Modified: #{record.last_modified.to_i - now.to_i}sec #{uri} #{record.encode.inspect}"
+          STDERR.puts "#{now.iso8601} info: strange LIRS: future Last-Modified: #{record.last_modified.to_i - now.to_i}sec #{uri} #{record.encode.inspect}\n"
         end
       }
     end
@@ -412,7 +507,7 @@ class Entry
     }
   rescue LIRS::Error
     # ignore invalid lirs file.
-    STDERR.puts "#{Time.now.iso8601} LIRS Error: #{uri}" if $VERBOSE
+    STDERR.puts "#{Time.now.iso8601} LIRS Error: #{uri}\n" if $VERBOSE
   end
 
   def path2pattern(*paths)
@@ -437,7 +532,8 @@ class Entry
   def decode_content_encoding(page, log)
     content = page
     if page.content_encoding.empty?
-      if /\A\x1f\x8b/ =~ content # gziped?
+      content = content.dup.force_encoding("ascii-8bit") if content.respond_to? :force_encoding
+      if /\A\x1f\x8b/n =~ content # gziped?
         begin
           content = content.decode_gzip
         rescue Zlib::Error
@@ -503,11 +599,11 @@ class Entry
           }
         end
       }
-      STDERR.puts "LIRS total: #{count_update} / #{count_interest} / #{count_all} - #{@config['URI']}" if $VERBOSE
+      STDERR.puts "LIRS total: #{count_update} / #{count_interest} / #{count_all} - #{@config['URI']}\n" if $VERBOSE
     rescue
       # External update information is a just hint.
       # So it is ignorable even if it has some trouble.
-      STDERR.puts "check_lirs: error on #{@config['URI']}: #$!"
+      STDERR.puts "check_lirs: error on #{@config['URI']}: #$!\n"
       #pp $!.backtrace
     end
   end
@@ -579,7 +675,7 @@ class Entry
       end
     }
 
-    STDERR.puts "HTML total: #{count_update} / #{count_interest} - #{@config['URI']}" if $VERBOSE
+    STDERR.puts "HTML total: #{count_update} / #{count_interest} - #{@config['URI']}\n" if $VERBOSE
   end
 
   def check_html(new_log=nil)
@@ -598,7 +694,7 @@ class Entry
     rescue
       # External update information is a just hint.
       # So it is ignorable even if it has some trouble.
-      STDERR.puts "check_html error on #{@config['URI']}: #$!"
+      STDERR.puts "check_html error on #{@config['URI']}: #$!\n"
     end
   end
 
@@ -886,8 +982,8 @@ class Entry
 
   def dump_filenames2
     log1, log2 = recent_log2
-    puts log1['content'].pathname if log1
-    puts log2['content'].pathname if log2
+    puts "#{log1['content'].pathname}\n" if log1
+    puts "#{log2['content'].pathname}\n" if log2
   end
 
   def diff_content
@@ -912,8 +1008,8 @@ class Entry
       text2 << [n.to_s, path]
     }
 
-    puts "checksum1: #{tree1.extract_text.to_s.sum} #{checksum_filter1.inspect} #{filename1}"
-    puts "checksum2: #{tree2.extract_text.to_s.sum} #{checksum_filter2.inspect} #{filename2}"
+    puts "checksum1: #{tree1.extract_text.to_s.sum} #{checksum_filter1.inspect} #{filename1}\n"
+    puts "checksum2: #{tree2.extract_text.to_s.sum} #{checksum_filter2.inspect} #{filename2}\n"
 
     [text1.length, text2.length].min.times {
       t1, p1 = text1.last
@@ -938,7 +1034,7 @@ class Entry
         pp [text1[i], text2[i]]
         num -= 1
         if num == 0
-          puts "..."
+          puts "...\n"
           break
         end
       end
@@ -994,7 +1090,7 @@ class Samidare
         stat1 = target.stat
         if nonblock
           unless target.flock(File::LOCK_EX | File::LOCK_NB)
-            STDERR.puts "fail to lock: #{filename}"
+            STDERR.puts "fail to lock: #{filename}\n"
             return
           end
         else
@@ -1043,7 +1139,7 @@ class Samidare
 
   def deep_freeze(o)
     objs = []
-    o = Marshal.load(Marshal.dump(o), lambda {|obj| objs << obj })
+    o = Marshal.load(Marshal.dump(o), lambda {|obj| objs << obj; obj })
     objs.each {|obj| obj.freeze }
     o
   end
@@ -1122,13 +1218,14 @@ class Samidare
   end
 
   def generate_output(data)
+    data = Presen.new(data)
     result = HTree.expand_template(GuessCharsetPathname.new(@opt_template), data, '')
     result << "\n" if /\n\z/ !~ result
     if @opt_output != '-'
       output_file(@opt_output, result)
       output_file(@opt_output + '.gz', result.encode_gzip)
     else
-      puts result
+      puts "#{result}\n"
     end
   end
 
@@ -1174,8 +1271,8 @@ class Samidare
       q.def_option('--verbose', '-v', 'verbose') { $VERBOSE = true }
       q.def_option('--no-check', '-n', 'don\'t check web') { @opt_dont_check = true }
       q.def_option('--force', '-f', 'force check (avoid timing control mechanism)') { @opt_force_check = true }
-      q.def_option('--output=filename', '-o', 'specify output file') {|filename| @opt_output = filename }
-      q.def_option('--output-lirs=filename', '-o', 'specify output file') {|filename| @opt_output_lirs = filename }
+      q.def_option('--output=filename', '-o', 'specify output html file') {|filename| @opt_output = filename }
+      q.def_option('--output-lirs=filename', 'specify output lirs file') {|filename| @opt_output_lirs = filename }
       q.def_option('--template=filename', '-T', 'specify template') {|filename| @opt_template = filename }
       q.def_option('--timing', '-t', 'show timings') { @opt_timing = true }
       q.def_option('--dump-config', 'dump flatten configuration') { @opt_dump_config = true }
@@ -1358,8 +1455,16 @@ class Samidare
           pp data
           return
         end
+        STDERR.print "generating output..." if $VERBOSE
+        t1 = Time.now
         generate_output(data)
+        t2 = Time.now
+        STDERR.puts " #{t2-t1}sec" if $VERBOSE
+        STDERR.print "generating lirs..." if $VERBOSE
+        t1 = Time.now
         generate_lirs(data)
+        t2 = Time.now
+        STDERR.puts " #{t2-t1}sec" if $VERBOSE
       end
     }
     #PP.pp(data, STDERR) if $VERBOSE
